/* stats.h
 * The loadable module that reads the timer and builds a stats array.
 *
 * Author:	Eyal Lebedinsky (eyal@ise.canberra.edu.au)
 * Last update:	28 June 1995
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/time.h>

#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/module.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

#include "stats.h"

#define ST_VERSION	"0.3"
#define ST_NAME		"stats"
#define ST_MAJOR	15
#define ST_BLOCK	1		/* 1=need to serialize access */

#define ST_MAJOR	15

#define PIT_MODE	0x43	/* io port for timer 0 */
#define PIT_COUNTER_0	0x40	/* io port for timer 0 */

#define GET_USERMEM(kernel,user) \
	get_usermem ((char *)&kernel, (char *)user, sizeof(kernel))

#define PUT_USERMEM(kernel,user) \
	put_usermem ((char *)&kernel, (char *)user, sizeof(kernel))

static char		kernel_version[] = UTS_RELEASE;
static int		read_semaphore = 0;
static int		mode = ST_DEF_MODE; /* 0=do_gettimeofday, 1=hw timer0 */
static unsigned long	count = ST_DEF_COUNT;
static unsigned long	array[ST_ARRAY_SIZE];
static unsigned long	total = 0;


/* Functions to read the hardware timer directly.
*/

#define LATCH		(1193180L/HZ)	/* timer 0 period */
#define DELTA_TIME(X,Y) ((X) - (Y) + ((X) >= (Y) ? 0 : LATCH))

/* Return the current value of timer 0. This is a 16 bit counter that starts
 * at LATCH and counts down to 0 then repeats.
*/
inline int 
get_timer0 (void)
{
	int	t;

	cli ();
	outb (0, PIT_MODE);
	t  = (int)inb (PIT_COUNTER_0);
	t += (int)inb (PIT_COUNTER_0) << 8;
	sti ();

	return (t);
}

static void
st_lock ()
{
	int	t;

	for (t = 1; t;) {
#if ST_BLOCK
		cli ();
#endif
		if (!(t = read_semaphore))
			read_semaphore = 1;
#if ST_BLOCK
		sti ();
#endif
	}
}

static void
st_unlock ()
{
	read_semaphore = 0;
}

static void
get_usermem (char *kernel, char *user, int size)
{
	verify_area (VERIFY_READ, (void *)user, size);
	st_lock ();
	while (size-- > 0)
		*kernel++ = get_fs_byte (user++);
	st_unlock ();
}

static void
put_usermem (char *kernel, char *user, int size)
{
	verify_area (VERIFY_WRITE, (void *)user, size);
	st_lock ();
	while (size-- > 0)
		put_fs_byte (*kernel++, user++);
	st_unlock ();
}

static int 
st_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	unsigned long arg)
{
	int		ret;
	int		tmode;

	ret = 0;
	switch (cmd) {
	case ST_SET_COUNT:
		GET_USERMEM (count, arg);
		break;
	case ST_GET_COUNT:
		PUT_USERMEM (count, arg);
		break;
	case ST_SET_MODE:
		GET_USERMEM (tmode, arg);
		if (tmode > ST_MODE_MAX)
			ret = -EINVAL;
#ifdef ST_NO_TOD
		else if (ST_MODE_TOD == tmode)
			ret = -EINVAL;
#endif
		else {
			st_lock ();
			mode = tmode;
			st_unlock ();
		}
		break;
	case ST_GET_MODE:
		PUT_USERMEM (mode, arg);
		break;
	case ST_CLEAR:
		st_lock ();
		memset (array, 0, sizeof (array));
		total = 0;
		st_unlock ();
		break;
	case ST_GET_TOTAL:
		PUT_USERMEM (total, arg);
		break;
	default:
		ret = -EINVAL;
		break;
	}
	return (ret);
}

static int
st_update (void)
{
	int		i;
	long		t;
	long		offset;
	int		t0, t1;
#ifndef ST_NO_TOD
	struct timeval	tv0, tv1;
#endif

	if (ST_MODE_TIMER == mode)
		t0 = get_timer0 ();
#ifndef ST_NO_TOD
	else
		do_gettimeofday (&tv0);
#endif
	for (i = 0; i < count; ++i) {
		if (ST_MODE_TIMER == mode) {
			t1 = get_timer0 ();
			t = DELTA_TIME (t0, t1);
			t0 = t1;
		}
#ifndef ST_NO_TOD
		else {
			do_gettimeofday (&tv1);
			t = (tv1.tv_sec - tv0.tv_sec) * 1000000L +
				(tv1.tv_usec - tv0.tv_usec);
			tv0 = tv1;
		}
#endif
		if (t < 0)
			t = 0;
		for (offset = 0; t >= ST_ARRAY_UNIT;) {
			t /= 10;
			offset += ST_ARRAY_UNIT;
			if (offset >= ST_ARRAY_SIZE) {
				offset = 0;
				t = ST_ARRAY_SIZE - 1;
				break;
			}
		}
		++array[offset+t];
		++total;
	}

	return (0);
}

/* The open will verify that this joystick exists even if it was missing at
 * init time. If the stick is disconnected later then timeouts will happen
 * on every read until the next open. This allows for intermittent failures.
*/
static int
st_open (struct inode *inode, struct file *file)
{
	if (MINOR (inode->i_rdev))
		return -ENODEV;

	st_lock ();

	count = ST_DEF_COUNT;
	mode  = ST_DEF_MODE;

	st_unlock ();

	MOD_INC_USE_COUNT;

	return (0);
}

static void
st_release (struct inode *inode, struct file *file)
{
	st_lock ();
	inode->i_atime = CURRENT_TIME;
	st_unlock ();

	MOD_DEC_USE_COUNT;
}

static int
st_read (struct inode *inode, struct file *file, char *buf, int size)
{
	if (size > sizeof (array))
		return -EOVERFLOW;

	st_lock ();

	inode->i_atime = CURRENT_TIME;
	st_update ();

	st_unlock ();

	put_usermem ((char *)array, buf, size);

	return (size);
}


static struct file_operations st_fops = {
	NULL,			/* st_lseek	*/
	st_read,		/* st_read	*/
	NULL,			/* st_write	*/
	NULL,			/* st_readaddr	*/
	NULL,			/* st_select	*/
	st_ioctl,		/* st_ioctl	*/
	NULL,			/* st_mmap	*/
	st_open,		/* st_open	*/
	st_release,		/* st_release	*/
	NULL			/* st_sync	*/
};

#ifdef __cplusplus
extern "C" {
#endif

int
init_module (void)
{
	printk ("%s: %s %s %s MAJOR=%d\n",
		ST_NAME, ST_VERSION, __DATE__, __TIME__, ST_MAJOR);

	if (register_chrdev (ST_MAJOR, ST_NAME, &st_fops)) {
		printk ("%s: unable to get MAJOR=%d\n", ST_NAME,
			ST_MAJOR);
		return (1);
	}

	count = ST_DEF_COUNT;
	read_semaphore = 0;

	return (0);
}

void
cleanup_module (void)
{
	if (MOD_IN_USE)
		printk ("%s: device busy, remove delayed.\n", ST_NAME);
	else {
		if (unregister_chrdev (ST_MAJOR, ST_NAME) != 0)
			printk ("%s: cleanup_module failed.\n", ST_NAME);
		else
			printk ("%s: cleanup_module succeeded.\n", ST_NAME);
	}
}

#ifdef __cplusplus
}
#endif
