/* timestat.c
 *
 * A program to use module stats.o to measure timer stability.
 *
 * Author:	Eyal Lebedinsky (eyal@ise.canberra.edu.au)
 * Last change:	30 June 1995
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <time.h>
#include <sys/timeb.h>

#include "stats.h"

static char		version[] = "0.2";

#ifndef CONTROL_FILE
#define	CONTROL_FILE	"/var/run/timestat.lock"
#endif

static unsigned long	array[ST_ARRAY_SIZE];
static unsigned long	prev[ST_ARRAY_SIZE];
static int		fd = -1;

static unsigned long	count = ST_DEF_COUNT;
static int		mode = ST_DEF_MODE;
static unsigned long	repeat = 1000;
static unsigned long	snapshot = 0;
static int		daemon = 0;
static int		verbose = 0;
static int		xlog = 0;

static char		*start = NULL;

void
die (char *str, int exitcode)
{
	if (str)
		fprintf (stderr, "%s failed: %s\n",
			str, strerror (errno));
	if (start) {
		free (start);
		start = NULL;
	}
	if (fd >= 0) {
		close (fd);
		fd = -1;
	}
	exit (exitcode);
}

static unsigned long	time_scale[] = {
	1UL, 10UL, 100UL, 1000UL, 10000UL, 100000UL, 1000000UL
};

static unsigned long
real_time (int i)
{
	int	n;

	n = i / ST_ARRAY_UNIT;
	i -= n * ST_ARRAY_UNIT;
	return (time_scale[n] * i);
}

static int
array_time (unsigned long t)
{
	int	offset;

	for (offset = 0; t >= ST_ARRAY_UNIT;) {
		t /= 10;
		offset += ST_ARRAY_UNIT;
		if (offset >= ST_ARRAY_SIZE) {
			offset = 0;
			t = ST_ARRAY_SIZE - 1;
			break;
		}
	}
	return ((int)(offset+t));
}

static char *
time_now (void)
{
	time_t		tm;
	char		*now;

	tm = time (0);
	now = ctime (&tm);
	now[strlen (now) - 1] = '\0';		/* kill NewLine */
	return (now);
}

static void
dump_log (unsigned long i)
{
	int		j;
	unsigned long	l;

	for (l = 0, j = ST_ARRAY_SIZE; --j >= 0;)
		prev[j] = (l += array[j]);

	printf ("\nversion    %s %s %s\n", version, __DATE__, __TIME__);
	printf ("start      %s\n", start);
	printf ("now        %s\n", time_now ());
	printf ("mode       %s%s%s\n",
		ST_MODE_TIMER == mode ? "timer (0.838us)" : "tod (1us)",
		daemon ? " daemon" : "",
		verbose ? " verbose" : "");
	printf ("count      %10lu\n", count);
	if (repeat)
		printf ("repeat     %10lu\n", repeat);
	if (!repeat || i < repeat)
		printf ("repeated   %10lu\n", i);
	if (snapshot)
		printf ("snapshot   %10lu\n", snapshot);

	printf ("\n%10s %10s %10s\n", "delay", "count", "accum");
	for (j = 0; j < ST_ARRAY_SIZE; ++j) {
		if (array[j])
			printf ("%10lu %10lu %10lu\n", 
				real_time (j), array[j], prev[j]);
	}
	fflush (stdout);
	memcpy (prev, array, sizeof (prev));
}

static unsigned long
get_ulong (char *s, unsigned long *n)
{
	unsigned long	l;

	for (l = 0; *s; ++s) {
		switch (*s) {
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			l = 10UL * l + (*s - '0');
			break;
		case 'k':
		case 'K':
			l *= 1000UL;
			break;
		case 'm':
		case 'M':
			l *= 1000000UL;
			break;
		}
	}
	*n = l;
	return (0);
}

int
main (int argc, char *argv[])
{
	int		errs;
	unsigned long	i;
	int		j;
	unsigned long	l;
	FILE		*f;
	int		title = 0;

	errs = 0;
	xlog = array_time (1000UL);

	for (i = 1; i < argc; ++i) {
		if (!strcmp ("timer", argv[i]))
			mode = ST_MODE_TIMER;
		else if (!strcmp ("tod", argv[i]))
			mode = ST_MODE_TOD;
		else if ('c' == argv[i][0]) {
			if (get_ulong (argv[i]+1, &l) || l <= 0) {
				printf ("bad count '%s'\n", argv[i]);
				++errs;
			} else
				count = l;
		} else if ('r' == argv[i][0]) {
			if (get_ulong (argv[i]+1, &l)) {
				printf ("bad repeat '%s'\n", argv[i]);
				++errs;
			} else
				repeat = l;
		} else if ('s' == argv[i][0]) {
			if (get_ulong (argv[i]+1, &l)) {
				printf ("bad snapshot '%s'\n", argv[i]);
				++errs;
			} else
				snapshot = l;
		} else if ('x' == argv[i][0]) {
			if (get_ulong (argv[i]+1, &l)) {
				printf ("bad exception '%s'\n", argv[i]);
				++errs;
			} else
				xlog = array_time (l);
		} else if (!strcmp ("v", argv[i]))
			++verbose;
		else if (!strcmp ("d", argv[i]))
			++daemon;
		else {
			printf ("unknown argument '%s'\n", argv[i]);
			++errs;
		}
	}

	if (errs) {
		printf ("usage: %s [timer|tod] [rNNN] [cNNN] [xNNN] [v] [d]\n", 
			argv[0]);
		exit (1);
	}

	if ((fd = open ("/dev/test", O_RDONLY)) < 0)
		die ("open", -1);

	if (ioctl (fd, ST_SET_COUNT, &count) < 0)
		die ("ST_SET_COUNT", -1);

	if (ioctl (fd, ST_SET_MODE, &mode) < 0)
		die ("ST_SET_MODE", -1);

	if (ioctl (fd, ST_CLEAR, NULL) < 0)
		die ("ST_CLEAR", -1);

	if (daemon) {
		if (NULL != fopen (CONTROL_FILE, "r")) {
			printf ("control file '%s' already present\n", 
				CONTROL_FILE);
			++errs;
		} else if (NULL == (f = fopen (CONTROL_FILE, "w"))) {
			printf ("could not create control file '%s'\n", 
				CONTROL_FILE);
			++errs;
		} else
			fclose (f);
		if (errs) {
			errno = 0;
			die (NULL, -1);
		}
	}

	for (j = 0; j < ST_ARRAY_SIZE; ++j)
		prev[j] = 0;

	if (NULL == (start = strdup (time_now ()))) {
		printf ("could not get memory\n");
		errno = 0;
		die ("no memory", -1);
	}

	for (i = 0; !repeat || i < repeat;) {
		if (read (fd, array, sizeof (array)) != sizeof (array))
			die ("read", -1);
		if (xlog) {
			for (j = xlog; j < ST_ARRAY_SIZE; ++j) {
				if (array[j] != prev[j]) {
					if (!title) {
						title = 1;
						printf ("\n%10s %10s %s\n", 
							"exception", "count",
							"when");
					}
					printf ("%10lu %10lu %s\n",
						real_time (j), array[j],
						time_now ());
					fflush (stdout);
				}
			}
		}
		memcpy (prev, array, sizeof (prev));

		if (!(++i % 100)) {
			if (verbose) {
				if (!(i % (100*50)))
					fprintf (stderr, "\n%lu", i);
				fprintf (stderr, ".");
			}
			if (daemon) {
				if (NULL == (f = fopen (CONTROL_FILE, "r")))
					break;
				fclose (f);
			}
		}
		if (snapshot && !(i % snapshot)) {
			dump_log (i);
			title = 0;
		}
	}
	if (verbose)
		fprintf (stderr, "\n");

	dump_log (i);

	close (fd);
	fd = -1;

	exit (0);
	return (0);
}
